import dateutil.parser
import uuid
from functools import partial
from typing import Tuple, Dict, List, Union, Type
from marshmallow import Schema, validate
from marshmallow.fields import Nested, String, Str, Int, Integer, UUID, DateTime, Boolean, Bool, List as MarshmallowList


class Sort:
    def __init__(self, name, base_class: Type[Schema]):
        self._name = name
        self._base_class = base_class

    def __call__(self, include: tuple = None, exclude: tuple = None):
        base_class_fields = self._base_class().declared_fields
        sort_fields = {}
        if not include:
            include = base_class_fields.keys()
        if exclude:
            include = {k: v for k, v in include if k not in exclude}
        for field in base_class_fields:
            if field in include:
                sort_fields[field] = Int(validate=validate.OneOf([1, -1]))
        sort_class = type(self._name, (Schema,), sort_fields)
        return Nested(sort_class)


class Filter(Nested):
    """
    Type for define filterable fields
    """
    _MARSHMALLOW_MAP = {
        String: 'string',
        Str: 'string',
        Integer: 'number',
        Int: 'number',
        UUID: 'uuid',
        DateTime: 'datetime',
        Boolean: 'boolean',
        Bool: 'boolean',
        List: 'list'
    }

    def __init__(self, *args, **kwargs):
        self._filterable_fields = kwargs.get("filterable")
        self._entity_schema = args[0]
        self._current_field = None
        super().__init__(*args, **kwargs)

    def _serialize(self, nested_obj, attr, obj, **kwargs):
        result = {}
        special_serializers = {
            '$like': self._serialize_like,
            '$ilike': partial(self._serialize_like, case_insensitive=True),
            '$in': self._serialize_in
        }
        for field_name, field_value in nested_obj.items():
            if v := special_serializers.get(field_name, lambda x: None)(field_value):
                return v
            elif isinstance(field_value, dict):
                self._current_field = field_name
                result[field_name] = self._serialize(field_value, field_name, obj, **kwargs)
            else:
                try:
                    cls = self._entity_schema().declared_fields[attr].__class__
                except Exception:
                    cls = self._entity_schema().declared_fields[field_name].__class__

                result[field_name] = self._process_payload(
                    cls,
                    field_value
                )

        return result

    def _serialize_like(self, value, case_insensitive=False):
        result = {
            '$regex': self._process_payload(
                Str,
                value
            )
        }
        if case_insensitive:
            result["$options"] = "i"
        return result

    def _serialize_in(self, value):
        if isinstance(value, list):
            cls = self._entity_schema().declared_fields[self._current_field].__class__
            return {"$in": [self._process_payload(cls, v) for v in value]}
        return value

    @staticmethod
    def _process_payload(type, value):
        _map = {
            DateTime: dateutil.parser.parse,
            String: str,
            Str: str,
            Integer: int,
            Int: int,
            UUID: uuid.UUID,
            Boolean: bool,
            Bool: bool,
            List: list
        }
        return _map[type](value)

    def _jsonschema_type_mapping(self):
        properties, required_fields = self._build_schema_attributes()
        return {
            "type": "object",
            "properties": properties,
            "required": required_fields
        }

    def _build_schema_attributes(self) -> Tuple[Dict[str, Union[str, Dict]], List]:
        """Build attributes to build schema for filter field"""

        entity_schema = self._entity_schema

        if callable(entity_schema):
            entity_schema = entity_schema()

        properties = {}
        required_fields = []

        for name, cls in entity_schema.declared_fields.items():
            if name in self._filterable_fields:
                if entity_schema.fields.get(name).required:
                    required_fields.append(name)
                properties[name] = dict({
                    "$ref": "#/definitions/operators/{}".format(self._MARSHMALLOW_MAP.get(cls.__class__)),
                    "description": entity_schema.fields.get(name).metadata.get("description")
                })

        return properties, required_fields


class HttpIOSchema:

    def __init__(self, name, base_class: Type[Schema]):
        self._name = name
        self._base_class = base_class

    def _get_reduced_class(self, include: list = None, exclude: list = None):
        base_class_fields = self._base_class().declared_fields
        reduced_class_fields = {}
        if not include:
            include = base_class_fields.keys()
        if exclude:
            include = {k: v for k, v in include if k not in exclude}
        for field in base_class_fields:
            if field in include:
                reduced_class_fields[field] = base_class_fields[field]

        return type(
            self._name + 'Item',
            (Schema,),
            reduced_class_fields
        )


class IndexResponseSchema(HttpIOSchema):
    def __call__(self, fields: tuple = None):
        return type(f"{self._name}ResponseSchema", (Schema,), {
            "items": MarshmallowList(Nested(self._get_reduced_class(include=fields))),
            "total": Int()
        })


class IndexRequestSchema(HttpIOSchema):
    def __call__(self, filter_fields: tuple, sort_fields: tuple):
        return type(f"{self._name}RequestSchema", (Schema,), {
            "filter": Filter(
                self._base_class,
                filterable=filter_fields,
            ),
            "limit": Int(),
            "offset": Int(),
            "sort": Sort(f"{self._name}SortSchema", self._base_class)(
                include=sort_fields
            )
        })
